<?php

namespace Drupal\image_scale_fill\Plugin\ImageEffect;

use Drupal\Core\Image\ImageInterface;
use Drupal\image\ConfigurableImageEffectBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\image_scale_fill\Component\GdGaussianBlur;

/**
 * Provides a 'Scale and Fill Background Effect' for Image Styles.
 *
 * @ImageEffect(
 *   id = "scale_and_fill_background",
 *   label = @Translation("Scale and Fill Background"),
 *   description = @Translation("Scales an image to fit a specified area and fills the rest with a background effect like a blurred image or dominant color.")
 * )
 */
class ScaleFillBackgroundEffect extends ConfigurableImageEffectBase {

  /**
   * {@inheritdoc}
   */
  public function applyEffect(ImageInterface $image) {
    $width = $this->configuration['width'];
    $height = $this->configuration['height'];
    $background_type = $this->configuration['background_type'];
    $gradient_direction = $this->configuration['gradient_direction'];

    $background = imagecreatetruecolor($width, $height);

    switch ($background_type) {
      case 'blurred':
        $this->applyBlurredBackground($image, $background);
        break;
      case 'dominant_color':
        $this->applyDominantColorBackground($image, $background);
        break;
      case 'gradient':
        $this->applyGradientBackground($image, $background, $gradient_direction);
        break;
    }

    // Common final steps to place the original image centered on the background
    $this->placeCenterImage($image, $background);

    // Set the modified background as the final image resource
    $image->getToolkit()->setResource($background);
    return TRUE;
  }

  /**
   * Applies a blurred background.
   */
  private function applyBlurredBackground(ImageInterface $image, $background) {
    // Get original dimensions
    $original_resource = $image->getToolkit()->getResource();
    $original_width = imagesx($original_resource);
    $original_height = imagesy($original_resource);

    // Resize the image based on configuration
    $scaled_width = (int) ($original_width * $this->configuration['resize_ratio']);
    $scaled_height = (int) ($original_height * $this->configuration['resize_ratio']);

    // Resize the image
    $tinyImage = imagecreatetruecolor($scaled_width, $scaled_height);
    imagecopyresampled($tinyImage, $original_resource, 0, 0, 0, 0, $scaled_width, $scaled_height, $original_width, $original_height);

    // Apply Gaussian Blur
    $radius = $this->configuration['radius'];
    $sigma = $this->configuration['sigma'] ?? $radius * 2 / 3;
    $coeffs = GdGaussianBlur::gaussianCoeffs($radius, $sigma);
    GdGaussianBlur::applyCoeffs($tinyImage, $tinyImage, $coeffs, $radius, 'HORIZONTAL');
    GdGaussianBlur::applyCoeffs($tinyImage, $tinyImage, $coeffs, $radius, 'VERTICAL');

    // Scale the blurred image back up
    imagecopyresampled($background, $tinyImage, 0, 0, 0, 0, $this->configuration['width'], $this->configuration['height'], $scaled_width, $scaled_height);

    // Clean up the tiny image resource
    imagedestroy($tinyImage);
  }

  /**
   * Applies a background with the dominant color.
   */
  private function applyDominantColorBackground(ImageInterface $image, $background) {
    $dominantColor = $this->getDominantColor($image);

    // Fill the background
    $color = imagecolorallocate($background, $dominantColor['red'], $dominantColor['green'], $dominantColor['blue']);
    imagefill($background, 0, 0, $color);
  }

  /**
   * Applies a gradient background.
   */
  private function applyGradientBackground(ImageInterface $image, $background, $direction) {
    $dominantColor = $this->getDominantColor($image);
    $startColor = imagecolorallocate($background, $dominantColor['red'], $dominantColor['green'], $dominantColor['blue']);
    $endColor = imagecolorallocate($background, max(0, $dominantColor['red'] - 50), max(0, $dominantColor['green'] - 50), max(0, $dominantColor['blue'] - 50));

    $this->applyGradient($background, $startColor, $endColor, $direction, $this->configuration['width'], $this->configuration['height']);
  }

  /**
   * Interpolates colors for gradients.
   */
  private function interpolateColor($image, $startColor, $endColor, $factor) {
    $start_red = ($startColor >> 16) & 0xFF;
    $start_green = ($startColor >> 8) & 0xFF;
    $start_blue = $startColor & 0xFF;

    $end_red = ($endColor >> 16) & 0xFF;
    $end_green = ($endColor >> 8) & 0xFF;
    $end_blue = $endColor & 0xFF;

    $red = (int)($start_red + ($factor * ($end_red - $start_red)));
    $green = (int)($start_green + ($factor * ($end_green - $start_green)));
    $blue = (int)($start_blue + ($factor * ($end_blue - $start_blue)));

    return imagecolorallocate($image, $red, $green, $blue);
  }
  
  /**
   * Helper function to interpolate and apply gradient.
   */
  private function applyGradient($image, $startColor, $endColor, $direction, $width, $height) {
    if ($direction === 'vertical') {
      for ($y = 0; $y < $height; $y++) {
        $interpolatedColor = $this->interpolateColor($image, $startColor, $endColor, $y / $height);
        imageline($image, 0, $y, $width, $y, $interpolatedColor);
      }
    } else { // horizontal
      for ($x = 0; $x < $width; $x++) {
        $interpolatedColor = $this->interpolateColor($image, $startColor, $endColor, $x / $width);
        imageline($image, $x, 0, $x, $height, $interpolatedColor);
      }
    }
  }

  /**
   * Extracts the most dominant color from the image.
   */
  private function getDominantColor(ImageInterface $image) {
    $size = 100;  // Sample size for faster processing
    $resized = imagecreatetruecolor($size, $size);
    imagecopyresampled($resized, $image->getToolkit()->getResource(), 0, 0, 0, 0, $size, $size, $image->getWidth(), $image->getHeight());

    $colors = [];
    for ($x = 0; $x < $size; $x++) {
      for ($y = 0; $y < $size; $y++) {
        $rgb = imagecolorat($resized, $x, $y);
        $colors[] = $rgb;
      }
    }
    imagedestroy($resized);

    $colors = array_count_values($colors);
    arsort($colors);
    $mostFrequent = key($colors);

    return [
      'red' => ($mostFrequent >> 16) & 0xFF,
      'green' => ($mostFrequent >> 8) & 0xFF,
      'blue' => $mostFrequent & 0xFF,
    ];
  }

  /**
   * Places the original image centered over the background.
   */
  private function placeCenterImage(ImageInterface $image, $background) {
    $original_width = $image->getWidth();
    $original_height = $image->getHeight();
    $center_width = (int) ($original_width * min($this->configuration['width'] / $original_width, $this->configuration['height'] / $original_height));
    $center_height = (int) ($original_height * min($this->configuration['width'] / $original_width, $this->configuration['height'] / $original_height));
    $center_image = imagecreatetruecolor($center_width, $center_height);
    imagecopyresampled($center_image, $image->getToolkit()->getResource(), 0, 0, 0, 0, $center_width, $center_height, $original_width, $original_height);
    $center_x = ($this->configuration['width'] - $center_width) / 2;
    $center_y = ($this->configuration['height'] - $center_height) / 2;
    imagecopy($background, $center_image, $center_x, $center_y, 0, 0, $center_width, $center_height);
    imagedestroy($center_image);
  }

  /**
 * {@inheritdoc}
 */
  public function getSummary() {
    $summary = [
        '#markup' => $this->t('Scales and fills the background of an image according to specified settings.') . '<br>' .
        $this->t('Width: @width px', ['@width' => $this->configuration['width']]) . '<br>' .
        $this->t('Height: @height px', ['@height' => $this->configuration['height']]) . '<br>' .
        $this->t('Background Type: @type', ['@type' => $this->configuration['background_type']])
    ];

    // Append additional configuration details based on the background type
    switch ($this->configuration['background_type']) {
        case 'blurred':
            $summary['#markup'] .= '<br>' . $this->t('Blur Radius: @radius', ['@radius' => $this->configuration['radius']])
                                   . '<br>' . $this->t('Blur Sigma: @sigma', ['@sigma' => $this->configuration['sigma']])
                                   . '<br>' . $this->t('Resize Ratio: @resize_ratio', ['@resize_ratio' => $this->configuration['resize_ratio']]);
            break;
        case 'gradient':
            $summary['#markup'] .= '<br>' . $this->t('Gradient Direction: @direction', ['@direction' => $this->configuration['gradient_direction']]);
            break;
    }

    return $summary;
  }


  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'width' => 800,
      'height' => 600,
      'background_type' => 'blurred',
      'radius' => 10,
      'sigma' => 6.7,
      'resize_ratio' => 0.1,
      'gradient_direction' => 'vertical',  // Default gradient direction
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['width'] = [
      '#type' => 'number',
      '#title' => $this->t('Width'),
      '#default_value' => $this->configuration['width'],
      '#required' => TRUE,
    ];

    $form['height'] = [
      '#type' => 'number',
      '#title' => $this->t('Height'),
      '#default_value' => $this->configuration['height'],
      '#required' => TRUE,
    ];

    $form['background_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Background Type'),
      '#options' => [
        'blurred' => $this->t('Blurred'),
        'dominant_color' => $this->t('Dominant Color'),
        'gradient' => $this->t('Gradient')
      ],
      '#default_value' => $this->configuration['background_type'],
    ];

    $form['radius'] = [
      '#type' => 'number',
      '#title' => $this->t('Blur Radius'),
      '#description' => $this->t('Specifies the radius of the blur effect.'),
      '#default_value' => $this->configuration['radius'],
      '#min' => 1,
      '#max' => 50,
      '#required' => TRUE,
      '#states' => [
            'visible' => [
                ':input[name="data[background_type]"]' => ['value' => 'blurred']
            ]
       ],
    ];

    $form['sigma'] = [
      '#type' => 'number',
      '#title' => $this->t('Blur Sigma'),
      '#description' => $this->t('Determines the spread of the blur effect.'),
      '#default_value' => $this->configuration['sigma'],
      '#min' => 0.1,
      '#max' => 50,
      '#step' => 0.1,
      '#states' => [
            'visible' => [
                ':input[name="data[background_type]"]' => ['value' => 'blurred']
            ]
       ],
    ];

    $form['resize_ratio'] = [
      '#type' => 'number',
      '#title' => $this->t('Resize Ratio'),
      '#description' => $this->t('Percentage to scale the image before applying background effects.'),
      '#default_value' => $this->configuration['resize_ratio'],
      '#min' => 0.1,
      '#max' => 1,
      '#step' => 0.1,
      '#states' => [
            'visible' => [
                ':input[name="data[background_type]"]' => ['value' => 'blurred']
            ]
        ],
    ];

    $form['gradient_direction'] = [
      '#type' => 'select',
      '#title' => $this->t('Gradient Direction'),
      '#options' => [
        'vertical' => $this->t('Vertical'),
        'horizontal' => $this->t('Horizontal')
      ],
      '#default_value' => $this->configuration['gradient_direction'],
      '#states' => [
            'visible' => [
                ':input[name="data[background_type]"]' => ['value' => 'gradient']
            ]
        ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['width'] = $form_state->getValue('width');
    $this->configuration['height'] = $form_state->getValue('height');
    $this->configuration['background_type'] = $form_state->getValue('background_type');
    $this->configuration['radius'] = $form_state->getValue('radius');
    $this->configuration['sigma'] = $form_state->getValue('sigma');
    $this->configuration['resize_ratio'] = $form_state->getValue('resize_ratio');
    $this->configuration['gradient_direction'] = $form_state->getValue('gradient_direction');
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    return parent::calculateDependencies();
  }
}
